@@ -3,7 +3,7 @@
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001,2005 James Aylett
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013 Olly Betts
  * Copyright 2009 Frank J Bruzzaniti
  * Copyright 2012 Mihai Bivol
  *
@@ -44,6 +44,7 @@
 
 #include <xapian.h>
 
+#include "append_filename_arg.h"
 #include "atomparse.h"
 #include "commonhelp.h"
 #include "diritor.h"
@@ -114,63 +115,6 @@ p_notalnum(unsigned int c)
     return !isalnum(static_cast<unsigned char>(c));
 }
 
-static string
-shell_protect(const string & file)
-{
-    string safefile = file;
-#ifdef __WIN32__
-    bool need_to_quote = false;
-    for (string::iterator i = safefile.begin(); i != safefile.end(); ++i) {
-	unsigned char ch = *i;
-	if (!isalnum(ch) && ch < 128) {
-	    if (ch == '/') {
-		// Convert Unix path separators to backslashes.  C library
-		// functions understand "/" in paths, but external commands
-		// generally don't, and also may interpret a leading '/' as
-		// introducing a command line option.
-		*i = '\\';
-	    } else if (ch == ' ') {
-		need_to_quote = true;
-	    } else if (ch < 32 || strchr("<>\"|*?", ch)) {
-		// Check for invalid characters in the filename.
-		string m("Invalid character '");
-		m += ch;
-		m += "' in filename \"";
-		m += file;
-		m += '"';
-		throw m;
-	    }
-	}
-    }
-    if (safefile[0] == '-') {
-	// If the filename starts with a '-', protect it from being treated as
-	// an option by prepending ".\".
-	safefile.insert(0, ".\\");
-    }
-    if (need_to_quote) {
-	safefile.insert(0, "\"");
-	safefile += '"';
-    }
-#else
-    string::size_type p = 0;
-    if (!safefile.empty() && safefile[0] == '-') {
-	// If the filename starts with a '-', protect it from being treated as
-	// an option by prepending "./".
-	safefile.insert(0, "./");
-	p = 2;
-    }
-    while (p < safefile.size()) {
-	// Don't escape some safe characters which are common in filenames.
-	unsigned char ch = safefile[p];
-	if (!isalnum(ch) && strchr("/._-", ch) == NULL) {
-	    safefile.insert(p, "\\");
-	    ++p;
-	}
-	++p;
-    }
-#endif
-    return safefile;
-}
 
 static bool ensure_tmpdir() {
     if (!tmpdir.empty()) return true;
@@ -205,11 +149,13 @@ parse_pdfinfo_field(const char * p, const char * end, string & out, const char *
     parse_pdfinfo_field((P), (END), (OUT), FIELD":", CONST_STRLEN(FIELD) + 1)
 
 static void
-get_pdf_metainfo(const string & safefile, string &author, string &title,
+get_pdf_metainfo(const string & file, string &author, string &title,
 		 string &keywords)
 {
     try {
-	string pdfinfo = stdout_to_string("pdfinfo -enc UTF-8 " + safefile);
+	string cmd = "pdfinfo -enc UTF-8";
+	append_filename_argument(cmd, file);
+	string pdfinfo = stdout_to_string(cmd);
 
 	const char * p = pdfinfo.data();
 	const char * end = p + pdfinfo.size();
@@ -244,8 +190,9 @@ static void
 generate_sample_from_csv(const string & csv_data, string & sample, size_t sample_size)
 {
     // Add 3 to allow for a 4 byte utf-8 sequence being appended when
-    // output is sample_size - 1 bytes long.
-    sample.reserve(sample_size + 3);
+    // output is sample_size - 1 bytes long.  Use csv_data.size() if smaller
+    // since the user might reasonably set sample_size really high.
+    sample.reserve(min(sample_size + 3, csv_data.size()));
     size_t last_word_end = 0;
     bool in_space = true;
     bool in_quotes = false;
@@ -454,7 +401,7 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		skip(file, "required filter not installed", SKIP_VERBOSE_ONLY);
 		return;
 	    }
-	    cmd += shell_protect(file);
+	    append_filename_argument(cmd, file);
 	    try {
 		dump = stdout_to_string(cmd);
 	    } catch (ReadError) {
@@ -462,7 +409,7 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		return;
 	    }
 	} else if (mimetype == "text/html") {
-	    string text = d.file_to_string();
+	    const string & text = d.file_to_string();
 	    MyHtmlParser p;
 	    if (ignore_exclusions) p.ignore_metarobots();
 	    try {
@@ -504,15 +451,16 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		// FIXME: What charset is the file?  Look at contents?
 	    }
 	} else if (mimetype == "application/pdf") {
-	    string safefile = shell_protect(file);
-	    string cmd = "pdftotext -enc UTF-8 " + safefile + " -";
+	    string cmd = "pdftotext -enc UTF-8";
+	    append_filename_argument(cmd, file);
+	    cmd += " -";
 	    try {
 		dump = stdout_to_string(cmd);
 	    } catch (ReadError) {
 		skip_cmd_failed(file, cmd);
 		return;
 	    }
-	    get_pdf_metainfo(safefile, author, title, keywords);
+	    get_pdf_metainfo(file, author, title, keywords);
 	} else if (mimetype == "application/postscript") {
 	    // There simply doesn't seem to be a Unicode capable PostScript to
 	    // text converter (e.g. pstotext always outputs ISO-8859-1).  The
@@ -529,12 +477,16 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		skip(file, msg);
 		return;
 	    }
-	    string tmpfile = tmpdir + "/tmp.pdf";
-	    string safetmp = shell_protect(tmpfile);
-	    string cmd = "ps2pdf " + shell_protect(file) + " " + safetmp;
+	    string tmpfile = tmpdir;
+	    tmpfile += "/tmp.pdf";
+	    string cmd = "ps2pdf";
+	    append_filename_argument(cmd, file);
+	    append_filename_argument(cmd, tmpfile);
 	    try {
 		(void)stdout_to_string(cmd);
-		cmd = "pdftotext -enc UTF-8 " + safetmp + " -";
+		cmd = "pdftotext -enc UTF-8";
+		append_filename_argument(cmd, tmpfile);
+		cmd += " -";
 		dump = stdout_to_string(cmd);
 	    } catch (ReadError) {
 		skip_cmd_failed(file, cmd);
@@ -545,7 +497,7 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		throw;
 	    }
 	    try {
-		get_pdf_metainfo(safetmp, author, title, keywords);
+		get_pdf_metainfo(tmpfile, author, title, keywords);
 	    } catch (...) {
 		unlink(tmpfile.c_str());
 		throw;
@@ -555,8 +507,11 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		   startswith(mimetype, "application/vnd.oasis.opendocument."))
 	{
 	    // Inspired by http://mjr.towers.org.uk/comp/sxw2text
-	    string safefile = shell_protect(file);
-	    string cmd = "unzip -p " + safefile + " content.xml ; unzip -p " + safefile + " styles.xml";
+	    string cmd = "unzip -p";
+	    append_filename_argument(cmd, file);
+	    cmd += " content.xml ; unzip -p";
+	    append_filename_argument(cmd, file);
+	    cmd += " styles.xml";
 	    try {
 		OpenDocParser parser;
 		parser.parse_html(stdout_to_string(cmd));
@@ -566,7 +521,9 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		return;
 	    }
 
-	    cmd = "unzip -p " + safefile + " meta.xml";
+	    cmd = "unzip -p";
+	    append_filename_argument(cmd, file);
+	    cmd += " meta.xml";
 	    try {
 		MetaXmlParser metaxmlparser;
 		metaxmlparser.parse_html(stdout_to_string(cmd));
@@ -578,7 +535,8 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		// It's probably best to index the document even if this fails.
 	    }
 	} else if (mimetype == "application/vnd.ms-excel") {
-	    string cmd = "xls2csv -c' ' -q0 -dutf-8 " + shell_protect(file);
+	    string cmd = "xls2csv -c' ' -q0 -dutf-8";
+	    append_filename_argument(cmd, file);
 	    try {
 		dump = stdout_to_string(cmd);
 	    } catch (ReadError) {
@@ -587,7 +545,6 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	    }
 	} else if (startswith(mimetype, "application/vnd.openxmlformats-officedocument.")) {
 	    const char * args = NULL;
-	    string safefile = shell_protect(file);
 	    string tail(mimetype, 46);
 	    if (startswith(tail, "wordprocessingml.")) {
 		// unzip returns exit code 11 if a file to extract wasn't found
@@ -598,7 +555,11 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		// Extract the shared string table first, so our parser can
 		// grab those ready for parsing the sheets which will reference
 		// the shared strings.
-		string cmd = "unzip -p " + safefile + " xl/sharedStrings.xml ; unzip -p " + safefile + " xl/worksheets/sheet\\*.xml";
+		string cmd = "unzip -p";
+		append_filename_argument(cmd, file);
+		cmd += " xl/sharedStrings.xml ; unzip -p";
+		append_filename_argument(cmd, file);
+		cmd += " xl/worksheets/sheet\\*.xml";
 		try {
 		    XlsxParser parser;
 		    parser.parse_html(stdout_to_string(cmd));
@@ -619,7 +580,9 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	    }
 
 	    if (args) {
-		string cmd = "unzip -p " + safefile + args;
+		string cmd = "unzip -p";
+		append_filename_argument(cmd, file);
+		cmd += args;
 		try {
 		    XmlParser xmlparser;
 		    xmlparser.parse_html(stdout_to_string(cmd));
@@ -630,7 +593,9 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		}
 	    }
 
-	    string cmd = "unzip -p " + safefile + " docProps/core.xml";
+	    string cmd = "unzip -p";
+	    append_filename_argument(cmd, file);
+	    cmd += " docProps/core.xml";
 	    try {
 		MetaXmlParser metaxmlparser;
 		metaxmlparser.parse_html(stdout_to_string(cmd));
@@ -644,13 +609,14 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	} else if (mimetype == "application/x-abiword") {
 	    // FIXME: Implement support for metadata.
 	    XmlParser xmlparser;
-	    string text = d.file_to_string();
+	    const string & text = d.file_to_string();
 	    xmlparser.parse_html(text);
 	    dump = xmlparser.dump;
 	    md5_string(text, md5);
 	} else if (mimetype == "application/x-abiword-compressed") {
 	    // FIXME: Implement support for metadata.
-	    string cmd = "gzip -dc " + shell_protect(file);
+	    string cmd = "gzip -dc";
+	    append_filename_argument(cmd, file);
 	    try {
 		XmlParser xmlparser;
 		xmlparser.parse_html(stdout_to_string(cmd));
@@ -662,7 +628,8 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	} else if (mimetype == "text/rtf") {
 	    // The --text option unhelpfully converts all non-ASCII characters
 	    // to "?" so we use --html instead, which produces HTML entities.
-	    string cmd = "unrtf --nopict --html 2>/dev/null " + shell_protect(file);
+	    string cmd = "unrtf --nopict --html 2>/dev/null";
+	    append_filename_argument(cmd, file);
 	    MyHtmlParser p;
 	    p.ignore_metarobots();
 	    try {
@@ -681,7 +648,8 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	    // pod2text's output character set doesn't seem to be documented,
 	    // but from inspecting the source it looks like it's probably
 	    // iso-8859-1.
-	    string cmd = "pod2text " + shell_protect(file);
+	    string cmd = "pod2text";
+	    append_filename_argument(cmd, file);
 	    try {
 		dump = stdout_to_string(cmd);
 		convert_to_utf8(dump, "ISO-8859-1");
@@ -695,7 +663,8 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	    // actually better to use -e2 (ISO-8859-1) and then convert, so
 	    // let's do that for now until we handle Unicode "compatibility
 	    // decompositions".
-	    string cmd = "catdvi -e2 -s " + shell_protect(file);
+	    string cmd = "catdvi -e2 -s";
+	    append_filename_argument(cmd, file);
 	    try {
 		dump = stdout_to_string(cmd);
 		convert_to_utf8(dump, "ISO-8859-1");
@@ -704,8 +673,9 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		return;
 	    }
 	} else if (mimetype == "application/vnd.ms-xpsdocument") {
-	    string safefile = shell_protect(file);
-	    string cmd = "unzip -p " + safefile + " Documents/1/Pages/\\*.fpage";
+	    string cmd = "unzip -p";
+	    append_filename_argument(cmd, file);
+	    cmd += " Documents/1/Pages/\\*.fpage";
 	    try {
 		XpsXmlParser xpsparser;
 		dump = stdout_to_string(cmd);
@@ -745,7 +715,8 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 
 	    generate_sample_from_csv(dump, sample, sample_size);
 	} else if (mimetype == "application/vnd.ms-outlook") {
-	    string cmd = get_pkglibbindir() + "/outlookmsg2html " + shell_protect(file);
+	    string cmd = get_pkglibbindir() + "/outlookmsg2html";
+	    append_filename_argument(cmd, file);
 	    MyHtmlParser p;
 	    p.ignore_metarobots();
 	    try {
@@ -767,14 +738,16 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	    author = p.author;
 	} else if (mimetype == "image/svg+xml") {
 	    SvgParser svgparser;
-	    svgparser.parse_html(d.file_to_string());
+	    const string & text = d.file_to_string();
+	    md5_string(text, md5);
+	    svgparser.parse_html(text);
 	    dump = svgparser.dump;
 	    title = svgparser.title;
 	    keywords = svgparser.keywords;
 	    author = svgparser.author;
 	} else if (mimetype == "application/x-debian-package") {
-	    string cmd("dpkg-deb -f ");
-	    cmd += shell_protect(file);
+	    string cmd("dpkg-deb -f");
+	    append_filename_argument(cmd, file);
 	    cmd += " Description";
 	    const string & desc = stdout_to_string(cmd);
 	    // First line is short description, which we use as the title.
@@ -784,8 +757,8 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 		dump.assign(desc, idx + 1, string::npos);
 	    }
 	} else if (mimetype == "application/x-redhat-package-manager") {
-	    string cmd("rpm -q --qf '%{SUMMARY}\\n%{DESCRIPTION}' -p ");
-	    cmd += shell_protect(file);
+	    string cmd("rpm -q --qf '%{SUMMARY}\\n%{DESCRIPTION}' -p");
+	    append_filename_argument(cmd, file);
 	    const string & desc = stdout_to_string(cmd);
 	    // First line is summary, which we use as the title.
 	    string::size_type idx = desc.find('\n');
@@ -795,7 +768,9 @@ index_file(const string &file, const string &url, DirectoryIterator & d,
 	    }
 	} else if (mimetype == "application/atom+xml") {
 	    AtomParser atomparser;
-	    atomparser.parse_html(d.file_to_string());
+	    const string & text = d.file_to_string();
+	    md5_string(text, md5);
+	    atomparser.parse_html(text);
 	    dump = atomparser.dump;
 	    title = atomparser.title;
 	    keywords = atomparser.keywords;
@@ -1235,6 +1210,7 @@ main(int argc, char **argv)
     // Extensions to quietly ignore:
     mime_map["a"] = "ignore";
     mime_map["bin"] = "ignore";
+    mime_map["com"] = "ignore";
     mime_map["css"] = "ignore";
     mime_map["dat"] = "ignore";
     mime_map["db"] = "ignore";
@@ -1258,19 +1234,19 @@ main(int argc, char **argv)
     mime_map["tmp"] = "ignore";
     mime_map["ttf"] = "ignore";
 
-    commands["application/msword"] = "antiword -mUTF-8.txt ";
-    commands["application/vnd.ms-powerpoint"] = "catppt -dutf-8 ";
+    commands["application/msword"] = "antiword -mUTF-8.txt";
+    commands["application/vnd.ms-powerpoint"] = "catppt -dutf-8";
     // Looking at the source of wpd2html and wpd2text I think both output
     // UTF-8, but it's hard to be sure without sample Unicode .wpd files
     // as they don't seem to be at all well documented.
-    commands["application/vnd.wordperfect"] = "wpd2text ";
+    commands["application/vnd.wordperfect"] = "wpd2text";
     // wps2text produces UTF-8 output from the sample files I've tested.
-    commands["application/vnd.ms-works"] = "wps2text ";
+    commands["application/vnd.ms-works"] = "wps2text";
     // Output is UTF-8 according to "man djvutxt".  Generally this seems to
     // be true, though some examples from djvu.org generate isolated byte
     // 0x95 in a context which suggests it might be intended to be a bullet
     // (as it is in CP1250).
-    commands["image/vnd.djvu"] = "djvutxt ";
+    commands["image/vnd.djvu"] = "djvutxt";
 
     if (argc == 2 && strcmp(argv[1], "-v") == 0) {
 	// -v was the short option for --version in 1.2.3 and earlier, but
@@ -1294,33 +1270,35 @@ main(int argc, char **argv)
 "BASEDIR is the directory corresponding to URL (default: DIRECTORY).\n"
 "\n"
 "Options:\n"
-"  -d, --duplicates         set duplicate handling ('ignore' or 'replace')\n"
-"  -p, --no-delete          skip the deletion of documents corresponding to\n"
-"                           deleted files (--preserve-nonduplicates is a\n"
-"                           deprecated alias for --no-delete)\n"
-"  -e, --empty-docs=ARG     how to handle documents we extract no text from:\n"
-"                           ARG can be index, warn (issue a diagnostic and\n"
-"                           index), or skip.  (default: warn)\n"
-"  -D, --db=DATABASE        path to database to use\n"
-"  -U, --url=URL            base url BASEDIR corresponds to (default: /)\n"
-"  -M, --mime-type=EXT:TYPE map file extension EXT to MIME Content-Type TYPE\n"
-"                           (empty TYPE removes any MIME mapping for EXT)\n"
-"  -F, --filter=TYPE:CMD    process files with MIME Content-Type TYPE using\n"
-"                           command CMD, which should produce UTF-8 text on\n"
-"                           stdout e.g. -Fapplication/octet-stream:'strings -n8'\n"
-"  -l, --depth-limit=LIMIT  set recursion limit (0 = unlimited)\n"
-"  -f, --follow             follow symbolic links\n"
-"  -i, --ignore-exclusions  ignore meta robots tags and similar exclusions\n"
-"  -S, --spelling           index data for spelling correction\n"
-"  -m, --max-size           maximum size of file to index (in bytes or with a\n"
-"                           suffix of 'K'/'k', 'M'/'m', 'G'/'g')\n"
-"  -E, --sample-size=SIZE   sets the maximum number of bytes for the document\n"
-"                           text sample. (default SIZE = 512)\n"
-"  -v, --verbose            show more information about what is happening\n"
-"      --overwrite          create the database anew (the default is to update\n"
-"                           if the database already exists)" << endl;
-	    print_stemmer_help("     ");
-	    print_help_and_version_help("     ");
+"  -d, --duplicates          set duplicate handling ('ignore' or 'replace')\n"
+"  -p, --no-delete           skip the deletion of documents corresponding to\n"
+"                            deleted files (--preserve-nonduplicates is a\n"
+"                            deprecated alias for --no-delete)\n"
+"  -e, --empty-docs=ARG      how to handle documents we extract no text from:\n"
+"                            ARG can be index, warn (issue a diagnostic and\n"
+"                            index), or skip.  (default: warn)\n"
+"  -D, --db=DATABASE         path to database to use\n"
+"  -U, --url=URL             base url BASEDIR corresponds to (default: /)\n"
+"  -M, --mime-type=EXT:TYPE  map file extension EXT to MIME Content-Type TYPE\n"
+"                            (empty TYPE removes any MIME mapping for EXT)\n"
+"  -F, --filter=TYPE:CMD     process files with MIME Content-Type TYPE using\n"
+"                            command CMD, which should produce UTF-8 text on\n"
+"                            stdout e.g. -Fapplication/octet-stream:'strings -n8'\n"
+"  -l, --depth-limit=LIMIT   set recursion limit (0 = unlimited)\n"
+"  -f, --follow              follow symbolic links\n"
+"  -i, --ignore-exclusions   ignore meta robots tags and similar exclusions\n"
+"  -S, --spelling            index data for spelling correction\n"
+"  -m, --max-size            maximum size of file to index (in bytes or with a\n"
+"                            suffix of 'K'/'k', 'M'/'m', 'G'/'g')\n"
+"                            (default: unlimited)\n"
+"  -E, --sample-size=SIZE    maximum size for the document text sample\n"
+"                            (supports the same formats as --max-size).\n"
+"                            (default: 512)\n"
+"  -v, --verbose             show more information about what is happening\n"
+"      --overwrite           create the database anew (the default is to update\n"
+"                            if the database already exists)" << endl;
+	    print_stemmer_help("      ");
+	    print_help_and_version_help("      ");
 	    return 0;
 	}
 	case 'V':
@@ -1380,10 +1358,8 @@ main(int argc, char **argv)
 	}
 	case 'F': {
 	    const char * s = strchr(optarg, ':');
-	    if (s != NULL || !s[1]) {
-		string command(s + 1);
-		command += ' ';
-		commands[string(optarg, s - optarg)] = command;
+	    if (s != NULL && s[1]) {
+		commands[string(optarg, s - optarg)] = string(s + 1);
 	    } else {
 		cerr << "Invalid filter mapping '" << optarg << "'\n"
 			"Should be of the form TYPE:COMMAND, e.g. 'application/octet-stream:strings -n8'"
